#include <fcntl.h>
#include <sys/types.h>
#include <dirent.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>
#include <assert.h>

#include "bonnie.h"
#include "bon_io.h"

CFileOp::~CFileOp()
{
  close();
  if(m_name)
  {
    int len = strlen(m_name);
    for(int i = 0; i < m_num_files; i++)
    {
      sprintf(&m_name[len], ".%03d", i);
      unlink(m_name);
    }
    delete m_name;
  }
}

int CFileOp::seek(int offset, int whence)
{
  switch(whence)
  {
    case SEEK_SET:
      m_file_ind = offset / m_chunks_per_file;
      m_cur_pos = offset % m_chunks_per_file;
    break;
    case SEEK_CUR:
      m_cur_pos += offset;
      while(m_cur_pos < 0)
      {
        m_file_ind--;
        m_cur_pos += m_chunks_per_file;
      }
      while(m_cur_pos >= m_chunks_per_file)
      {
        m_file_ind++;
        m_cur_pos -= m_chunks_per_file;
      }
    break;
    default:
      fprintf(stderr, "unsupported seek option\n");
      return -1;
  }
  if(m_file_ind == m_num_files)
  {
    m_file_ind--;
    m_cur_pos += m_chunks_per_file;
  }
  else
  {
    if(m_file_ind < 0 || m_file_ind >= m_num_files
       || (m_file_ind == m_num_files - 1 && m_cur_pos >= m_last_file_chunks))
    {
      fprintf(stderr, "Bad seek offset\n");
      return -1;
    }
    int rc;
    if(m_fd)
      rc = lseek(m_fd[m_file_ind], m_cur_pos << ChunkBits, SEEK_SET);
    else
      rc = fseek(m_stream[m_file_ind], m_cur_pos << ChunkBits, SEEK_SET);

    if(rc == -1)
      fprintf(stderr, "Error in lseek to %d\n", (m_cur_pos << ChunkBits));
    else
      rc = 0;
    return rc;
  }
  return 0;
}

int CFileOp::read_block(PVOID buf)
{
  assert(m_fd);
  int rc = ::read(m_fd[m_file_ind], buf, Chunk);
  m_cur_pos++;
  if(m_cur_pos >= m_chunks_per_file)
  {
    if(seek(0, SEEK_CUR) == -1)
    {
      fprintf(stderr, "Error in seek(0)\n");
      return -1;
    }
  }
  if(rc == -1)
  {
    io_error("re-write read");
  }
  else if(rc != Chunk)
  {
    fprintf(stderr, "Can't read a full block, only got %d bytes.\n", rc);
    rc = -1;
  }
  return rc;
}

int CFileOp::read_block_getc(char *buf)
{
  assert(m_stream);
  int next;
  for(int i = 0; i < Chunk; i++)
  {
    if ((next = getc(m_stream[m_file_ind])) == EOF)
    {
      fprintf(stderr, "Can't getc(3)\n");
      return -1;
    }
    /* just to fool optimizers */
    buf[next]++;
  }

  m_cur_pos++;
  if(m_cur_pos >= m_chunks_per_file)
  {
    if(seek(0, SEEK_CUR) == -1)
      return -1;
  }
  return 0;
}

int CFileOp::write_block(CPCVOID buf)
{
  assert(m_fd);
  int rc = ::write(m_fd[m_file_ind], buf, Chunk);
  m_cur_pos++;
  if(m_cur_pos >= m_chunks_per_file)
  {
    if(seek(0, SEEK_CUR) == -1)
      return -1;
  }
  return rc;
}

int CFileOp::write_block_putc()
{
  assert(m_stream);
  for(int i = 0; i < Chunk; i++)
  {
    if (putc(i & 0x7f, m_stream[m_file_ind]) == EOF)
    {
      fprintf(stderr, "Can't putc() - disk full?\n");
      return -1;
    }
  }
  m_cur_pos++;
  if(m_cur_pos >= m_chunks_per_file)
  {
    if(seek(0, SEEK_CUR) == -1)
      return -1;
  }
  return 0;
}

int CFileOp::open(CPCCHAR basename, bool create, bool fopen)
{
  m_name = new char[strlen(basename) + 8];
  strcpy(m_name, basename);
  return reopen(create, fopen);
}

CFileOp::CFileOp()
 : m_chunks_per_file(Unit / Chunk * IOFileSize)
 , m_stream(NULL)
 , m_fd(NULL)
 , m_isopen(false), m_name(NULL)
{
}

void CFileOp::set_size(int size)
{
  m_total_chunks = Unit / Chunk * size;
  m_last_file_chunks = m_total_chunks % m_chunks_per_file;
  m_num_files = size / IOFileSize + 1;
  if(m_last_file_chunks == 0)
  {
    m_last_file_chunks = m_chunks_per_file;
    m_num_files--;
  }
}

typedef FILE * PFILE;

int CFileOp::reopen(bool create, bool fopen)
{
  int i;
  m_cur_pos = 0;
  m_file_ind = 0;

  if(m_isopen) close();

  m_isopen = true;
  if(fopen)
    m_stream = new PFILE[m_num_files];
  else
    m_fd = new int[m_num_files];
  for(i = 0; i < m_num_files; i++)
  {
    if(fopen)
      m_stream[i] = NULL;
    else
      m_fd[i] = -1;
  }
  int len = strlen(m_name);
  for(i = 0; i < m_num_files; i++)
  {
    sprintf(&m_name[len], ".%03d", i);
    if(m_open(m_name, i, create))
      return true;
  }
  m_name[len] = '\0';
  return 0;
}

int CFileOp::m_open(CPCCHAR basename, int ind, bool create)
{
  if(create)
  { /* create from scratch */
    unlink(basename);
    if(m_fd)
      m_fd[ind] = ::open(basename, O_RDWR | O_CREAT | O_EXCL, S_IRUSR | S_IWUSR);
    else
      m_stream[ind] = fopen(basename, "w+");
  } /* create from scratch */
  else
  {
    if(m_fd)
      m_fd[ind] = ::open(basename, O_RDWR, S_IRUSR | S_IWUSR);
    else
      m_stream[ind] = fopen(basename, "r+");
  }

  if( (m_fd && m_fd[ind] == -1) || (m_stream && m_stream[ind] == NULL) )
  {
    fprintf(stderr, "Can't open file %s\n", basename);
    return -1;
  }
  return 0;
}

void CFileOp::close()
{
  if(!m_isopen)
    return;
  for(int i = 0; i < m_num_files; i++)
  {
    if(m_stream) fclose(m_stream[i]);
    if(m_fd) ::close(m_fd[i]);
  }
  m_isopen = false;
  delete m_fd;
  delete m_stream;
  m_fd = NULL;
  m_stream = NULL;
}


/*
 * Do a typical-of-something random I/O.  Any serious application that
 *  has a random I/O bottleneck is going to be smart enough to operate
 *  in a page mode, and not stupidly pull individual words out at
 *  odd offsets.  To keep the cache from getting too clever, some
 *  pages must be updated.  However an application that updated each of
 *  many random pages that it looked at is hard to imagine.  
 * However, it would be wrong to put the update percentage in as a
 *  parameter - the effect is too nonlinear.  Need a profile
 *  of what Oracle or Ingres or some such actually does.
 * Be warned - there is a *sharp* elbow in this curve - on a 1-Mb file,
 *  most substantial unix systems show >2000 random I/Os per second -
 *  obviously they've cached the whole thing and are just doing buffer
 *  copies.  
 */
int 
CFileOp::doseek(long where, bool update)
{
  assert(m_fd);
  int   buf[Chunk / IntSize];
  int   size;

  if (seek(where, SEEK_SET) == -1)
    return io_error("lseek in doseek");
  if ((size = read_block(PVOID(buf))) == -1)
    return io_error("read in doseek");

  /* every so often, update a block */
  if (update)
  { /* update this block */

    /* touch a word */
    buf[(int(random()) % (size/IntSize - 2)) + 1]--;
    if(seek(where, SEEK_SET) == -1)
      return io_error("lseek in doseek update");
    if (write_block(PVOID(buf)) == -1)
      return io_error("write in doseek");
  } /* update this block */
  return 0;
}

